@@ -9,25 +9,19 @@
 from __future__ import annotations
 
 import asyncio
-import json
-import math
 import pathlib
 import signal
-import warnings
 from functools import wraps
-from tempfile import NamedTemporaryFile
 
-from typing import Annotated, Optional
+from typing import Annotated, Optional, cast
 
 import typer
 from rich.console import Console
 from typer import Argument, Option
 from typer.core import TyperGroup
 
-from lvmcryo import __version__
-from lvmcryo.config import Actions, Config, InteractiveMode, NotificationLevel
-from lvmcryo.runner import post_fill_tasks
-from lvmcryo.tools import add_json_handler
+from lvmcryo.config import Actions, InteractiveMode, NotificationLevel
+from lvmcryo.tools import DBHandler
 
 
 LOCKFILE = pathlib.Path("/data/lvmcryo.lock")
@@ -75,6 +69,7 @@ cli = typer.Typer(
 @cli.command("ln2")
 @cli_coro()
 async def ln2(
+    ctx: typer.Context,
     #
     # Arguments
     #
@@ -93,14 +88,23 @@ async def ln2(
     #
     # General options
     #
+    profile: Annotated[
+        Optional[str],
+        Option(
+            "--profile",
+            "-p",
+            envvar="LVMCRYO_PROFILE",
+            help="Profile to use. A list of valid profiles and parameters "
+            "can be printed with lvmcryo list-profiles.",
+        ),
+    ] = None,
     config_file: Annotated[
         Optional[pathlib.Path],
         Option(
             dir_okay=False,
             exists=True,
             envvar="LVMCRYO_CONFIG_FILE",
-            help="The configuration file to use to set the default values. "
-            "Defaults to the internal configuration file.",
+            help="The configuration file. Defaults to the internal configuration file.",
         ),
     ] = None,
     dry_run: Annotated[
@@ -157,10 +161,22 @@ async def ln2(
         bool,
         Option(
             " /--no-use-thermistors",
+            envvar="LVMCRYO_USE_THERMISTORS",
             help="Use thermistor values to determine purge/fill time.",
             rich_help_panel="Purge and fill options",
         ),
     ] = True,
+    require_all_thermistors: Annotated[
+        bool,
+        Option(
+            "--require-all-thermistors",
+            envvar="LVMCRYO_REQUIRE_ALL_THERMISTORS",
+            help="If set, waits until all thermistors have activated before closing "
+            "any of the valves. This prevents overpressures in the cryostats when only "
+            "some of the valves are open. Ignore if --no-use-thermistors is used. ",
+            rich_help_panel="Purge and fill options",
+        ),
+    ] = False,
     check_pressures: Annotated[
         bool,
         Option(
@@ -329,11 +345,14 @@ async def ln2(
         Option(
             "--write-json",
             envvar="LVMCRYO_WRITE_JSON",
-            help="Saves the log in JSON format. Uses the same path as the file log."
+            help="Saves the log in JSON format. Uses the same path as the file log. "
             "Ignored if --write-log is not passed.",
             rich_help_panel="Logging",
         ),
     ] = True,
+    #
+    # Post-fill data options
+    #
     write_data: Annotated[
         bool,
         Option(
@@ -341,25 +360,27 @@ async def ln2(
             envvar="LVMCRYO_WRITE_DATA",
             help="Saves cryostat pressures and temperatures taken during purge/fill "
             "to a parquet file.",
-            rich_help_panel="Logging",
+            rich_help_panel="Post-fill data logging",
         ),
     ] = False,
     data_path: Annotated[
         Optional[pathlib.Path],
         Option(
+            envvar="LVMCRYO_DATA_PATH",
             exists=False,
             help="Path where to save the data. Implies --write-data. "
             "Defaults to a path relative to the log path.",
-            rich_help_panel="Logging",
+            rich_help_panel="Post-fill data logging",
         ),
     ] = None,
     data_extra_time: Annotated[
         float,
         Option(
+            envvar="LVMCRYO_DATA_EXTRA_TIME",
             help="Additional time to take cryostat data after the "
             "action has been completed. The command will not complete until "
             "data collection has finished.",
-            rich_help_panel="Logging",
+            rich_help_panel="Post-fill data logging",
         ),
     ] = 60,
 ):
@@ -373,15 +394,23 @@ async def ln2(
     """
 
     from logging import FileHandler
+    from tempfile import NamedTemporaryFile
 
     from rich.prompt import Confirm
 
     from sdsstools.logger import get_logger
 
+    from lvmcryo import __version__
+    from lvmcryo.config import Config
     from lvmcryo.handlers.ln2 import LN2Handler, get_now
     from lvmcryo.notifier import Notifier
-    from lvmcryo.runner import ln2_runner
-    from lvmcryo.tools import LockExistsError, ensure_lock
+    from lvmcryo.runner import ln2_runner, post_fill_tasks
+    from lvmcryo.tools import (
+        LockExistsError,
+        add_json_handler,
+        ensure_lock,
+    )
+    from lvmcryo.validate import validate_fill
 
     # Create log here and use its console for stdout.
     log = get_logger("lvmcryo", use_rich_handler=True)
@@ -395,7 +424,6 @@ async def ln2(
 
     json_path: pathlib.Path | None = None
     json_handler: FileHandler | None = None
-    log_data: list[dict] | None = None
     images: dict[str, pathlib.Path | None] = {}
 
     if action == Actions.abort:
@@ -414,6 +442,8 @@ async def ln2(
             cameras=cameras or [],
             config_file=config_file,
             dry_run=dry_run,
+            clear_lock=clear_lock,
+            with_traceback=with_traceback,
             interactive=interactive,
             no_prompt=no_prompt,
             notify=notify,
@@ -421,6 +451,7 @@ async def ln2(
             email=email,
             email_level=email_level,
             use_thermistors=use_thermistors,
+            require_all_thermistors=require_all_thermistors,
             check_pressures=check_pressures,
             check_temperatures=check_temperatures,
             max_pressure=max_pressure,
@@ -440,6 +471,11 @@ async def ln2(
             data_path=data_path,
             data_extra_time=data_extra_time,
             version=__version__,
+            profile=profile,
+            # We cannot pass the context directly so we pass a dict of the
+            # origin of each parameter to reject profile parameters that
+            # have been manually defined.
+            param_source={pp: ctx.get_parameter_source(pp) for pp in ctx.params},
         )
     except ValueError as err:
         err_console.print(f"[red]Error parsing configuration:[/] {err}")
@@ -448,7 +484,7 @@ async def ln2(
     if config.write_log and config.log_path:
         log.start_file_logger(str(config.log_path))
 
-        if write_json:
+        if config.write_json:
             json_path = config.log_path.with_suffix(".json")
             json_handler = add_json_handler(log, json_path)
 
@@ -476,8 +512,8 @@ async def ln2(
     if not config.notify:
         log.debug("Notifications are disabled and will not be emitted.")
 
-    if config_file is not None:
-        log.info(f"Using configuration file: {config_file!s}")
+    if config.config_file is not None:
+        log.info(f"Using configuration file: {config.config_file!s}")
 
     if not config.no_prompt:
         stdout_console.print(f"Action {config.action.value} will run with:")
@@ -500,20 +536,32 @@ async def ln2(
         alerts_route=config.internal_config["api_routes"]["alerts"],
     )
 
-    if LOCKFILE.exists() and clear_lock:
+    if LOCKFILE.exists() and config.clear_lock:
         log.warning("Lock file exists. Removing it because --clear-lock.")
         LOCKFILE.unlink()
 
+    db_handler = DBHandler(action, handler, config, json_handler=json_handler)
+    record_pk = await db_handler.write(complete=False)
+    if record_pk:
+        log.debug(f"Record {record_pk} created in the database.")
+
     try:
         with ensure_lock(LOCKFILE):
+            # Calculate the expected maximum run time.
+            max_time: float = 2 * 3600  # It should never take longer than two hours.
+            if config.max_purge_time is not None and config.max_fill_time is not None:
+                max_time = config.max_purge_time + config.max_fill_time + 300.0
+
             # Run worker.
-            await ln2_runner(handler, config, notifier)
+            await asyncio.wait_for(
+                ln2_runner(handler, config, notifier, db_handler=db_handler),
+                timeout=max_time,
+            )
 
             # Check handler status.
             if handler.failed:
                 raise RuntimeError(
-                    "No exceptions were raised but the LN2 "
-                    "handler reports a failure."
+                    "No exceptions were raised but the LN2 handler reports a failure."
                 )
 
     except LockExistsError:
@@ -523,7 +571,7 @@ async def ln2(
             "the lock."
         )
 
-        if notify:
+        if config.notify:
             log.warning("Sending failure notifications.")
             await notifier.notify_after_fill(
                 False,
@@ -547,7 +595,7 @@ async def ln2(
         handler.failed = True
 
         error = err
-        if with_traceback:
+        if config.with_traceback:
             raise
 
         raise typer.Exit(1)
@@ -556,51 +604,60 @@ async def ln2(
         log.info(f"LN2 {action.value} completed successfully.")
 
     finally:
+        # At this point all the valves are closed so we can remove the signal handlers.
+        for signame in ("SIGINT", "SIGTERM"):
+            asyncio.get_running_loop().remove_signal_handler(getattr(signal, signame))
+
         handler.event_times.end_time = get_now()
         await handler.clear()
 
         log.info(f"Event times:\n{handler.event_times.model_dump_json(indent=2)}")
 
         if not skip_finally:
-            if json_handler and json_path:
-                json_handler.flush()
-                with json_path.open("r") as ff:
-                    log_data = [json.loads(line) for line in ff.readlines()]
+            # Do a quick update of the DB record since post_fill_tasks() may
+            # block for a long time.
+            await db_handler.write(error=error)
 
-            configuration_json = config.model_dump() | {
-                valve: valve_model.model_dump()
-                for valve, valve_model in config.valve_info.items()
-            }
-            record_pk, plot_paths = await post_fill_tasks(
+            plot_paths = await post_fill_tasks(
                 handler,
                 notifier=notifier,
                 write_data=config.write_data,
                 data_path=config.data_path,
                 data_extra_time=config.data_extra_time if error is None else None,
                 api_data_route=config.internal_config["api_routes"]["fill_data"],
-                write_to_db=True,
-                api_db_route=config.internal_config["api_routes"]["register_fill"],
-                db_extra_payload={
-                    "error": str(error) if error is not None else None,
-                    "action": action.value,
-                    "log_file": str(config.log_path) if config.log_path else None,
-                    "json_file": str(json_path) if json_path and write_json else None,
-                    "log_data": log_data,
-                    "configuration": configuration_json,
-                    "valve_times": handler.get_valve_times(as_string=True),
-                },
             )
 
-            images = {
-                "pressure": plot_paths.get("pressure_png", None),
-                "temps": plot_paths.get("temps_png", None),
-                "thermistors": plot_paths.get("thermistors_png", None),
-            }
+            if (
+                config.write_data
+                and config.data_path
+                and config.data_path.exists()
+                and not error
+            ):
+                validate_failed, validate_error = validate_fill(
+                    handler,
+                    config,
+                    log=log,
+                )
+                if validate_failed and error is None:
+                    await notifier.post_to_slack(
+                        "Fill validation failed. Check the log for details.",
+                        level=NotificationLevel.error,
+                    )
+                    handler.failed = True
+                    error = validate_error
+                elif not validate_failed:
+                    log.info("Fill validation completed successfully.")
 
-            if record_pk:
-                log.debug(f"Record {record_pk} created in the database.")
+            log.info("Writing fill metadata to database.")
+            await db_handler.write(complete=True, plot_paths=plot_paths, error=error)
+
+            if config.notify:
+                images = {
+                    "pressure": plot_paths.get("pressure_png", None),
+                    "temps": plot_paths.get("temps_png", None),
+                    "thermistors": plot_paths.get("thermistors_png", None),
+                }
 
-            if notify:
                 if error:
                     log.warning("Sending failure notifications.")
                     await notifier.notify_after_fill(
@@ -650,6 +707,31 @@ async def _close_valves_helper():
     return typer.Exit(0)
 
 
+@cli.command("list-profiles")
+def list_profiles(
+    config_file: Annotated[
+        Optional[pathlib.Path],
+        Option(
+            dir_okay=False,
+            exists=True,
+            envvar="LVMCRYO_CONFIG_FILE",
+            help="The configuration file. Defaults to the internal configuration file.",
+        ),
+    ] = None,
+):
+    """Lists the available profiles."""
+
+    from lvmcryo.config import get_internal_config
+
+    internal_config = get_internal_config(config_file=config_file)
+    profiles = internal_config["profiles"]
+
+    for profile in profiles:
+        info_console.print(profile, style="bold")
+        info_console.print(profiles[profile])
+        print()
+
+
 @cli.command("close-valves")
 @cli_coro()
 async def close_valves():
@@ -678,6 +760,13 @@ async def ion(
             show_default=False,
         ),
     ] = None,
+    skip_pressure_check: Annotated[
+        bool,
+        Option(
+            "--skip-pressure-check",
+            help="Skips the pressure check when turning the ion pump on.",
+        ),
+    ] = False,
 ):
     """Controls the ion pumps.
 
@@ -686,7 +775,13 @@ async def ion(
 
     """
 
+    import math
+    import warnings
+
     from lvmopstools.devices.ion import read_ion_pumps, toggle_ion_pump
+    from lvmopstools.devices.specs import Spectrographs, spectrograph_pressures
+
+    from lvmcryo.config import get_internal_config
 
     cameras = list(map(lambda x: x.lower(), cameras)) if cameras else ["all"]
 
@@ -726,8 +821,29 @@ async def ion(
 
     error: bool = False
 
+    if cameras == ["all"]:
+        config = get_internal_config()
+        cameras = config["defaults"]["cameras"]
+
+    if cameras is None:
+        err_console.print("[red]No cameras were specified.[/]")
+        raise typer.Exit(1)
+
     for camera in cameras:
         try:
+            if on is True and not skip_pressure_check:
+                spec = f"sp{camera[-1]}"
+                pressures = await spectrograph_pressures(cast(Spectrographs, spec))
+                cam_pressure = pressures.get(camera, None)
+                if cam_pressure is None:
+                    raise ValueError(f"Could not get pressure for {camera!r}.")
+
+                if cam_pressure > 1e-4:
+                    raise ValueError(
+                        f"Camera {camera} has a pressure of {cam_pressure:.2e} "
+                        "Torr. If you want to turn it on use --skip-pressure-check."
+                    )
+
             await toggle_ion_pump(camera, on)
         except Exception as err:
             info_console.print(
