@@ -18,18 +18,18 @@ from typing import TYPE_CHECKING, Any, Optional
 from rich.progress import TaskID
 
 from lvmopstools.clu import CluClient
+from lvmopstools.retrier import Retrier
 
 from lvmcryo.config import get_internal_config
 from lvmcryo.handlers.thermistor import ThermistorHandler
-from lvmcryo.tools import (
-    cancel_task,
-    get_fake_logger,
-)
+from lvmcryo.tools import cancel_task, get_fake_logger, ln2_estops
 
 
 if TYPE_CHECKING:
     from sdsstools.configuration import Configuration
 
+    from lvmcryo.tools import TimerProgressBar
+
 
 __all__ = [
     "ValveHandler",
@@ -40,6 +40,7 @@ __all__ = [
 ]
 
 
+@Retrier(max_attempts=3, delay=1, timeout=10)
 async def outlet_info(actor: str, outlet: str) -> dict[str, Any]:
     """Retrieves outlet information from the NPS."""
 
@@ -51,6 +52,7 @@ async def outlet_info(actor: str, outlet: str) -> dict[str, Any]:
     return cmd.replies.get("outlet_info")
 
 
+@Retrier(max_attempts=3, delay=1, timeout=30)
 async def valve_on_off(
     actor: str,
     outlet_name: str,
@@ -87,6 +89,11 @@ async def valve_on_off(
 
     """
 
+    # Check if the LN2 e-stops are active. If so we cannot operate the valves
+    # because the NPSs that control them will be powered off.
+    if await ln2_estops():
+        raise RuntimeError("Cannot operate LN2 valves: e-stops are active.")
+
     is_script: bool = False
 
     if on is True and isinstance(timeout, (int, float)) and use_script is True:
@@ -118,6 +125,7 @@ async def valve_on_off(
     return
 
 
+@Retrier(max_attempts=3, delay=1, timeout=10)
 async def cancel_nps_threads(actor: str, thread_id: int | None = None):
     """Cancels a script thread in an NPS.
 
@@ -136,6 +144,7 @@ async def cancel_nps_threads(actor: str, thread_id: int | None = None):
         await client.send_command(actor, command_string)
 
 
+@Retrier(max_attempts=3, delay=1, timeout=60)
 async def close_all_valves(config: Configuration | None = None, dry_run: bool = False):
     """Closes all the outlets."""
 
@@ -155,10 +164,6 @@ async def close_all_valves(config: Configuration | None = None, dry_run: bool =
     )
 
 
-if TYPE_CHECKING:
-    from lvmcryo.tools import TimerProgressBar
-
-
 @dataclass
 class ValveHandler:
     """Handles a valve, including opening and closing, timeouts, and thermistors.
@@ -224,6 +229,7 @@ class ValveHandler:
         min_open_time: float = 0.0,
         max_open_time: float | None = None,
         use_thermistor: bool = True,
+        close_on_active: bool = True,
     ):
         """Starts a fill.
 
@@ -238,6 +244,9 @@ class ValveHandler:
             Whether to use the thermistor to close the valve. If ``True`` and
             ``fill_time`` is not ``None``, ``fill_time`` become the maximum
             open time.
+        close_on_active
+            If ``use_thermistor=True``, closes the valve when the thermistor becomes
+            active. Otherwise just blocks while the thermistor is inactive.
 
         """
 
@@ -257,6 +266,7 @@ class ValveHandler:
 
         if use_thermistor and self.thermistor_info:
             thermistor_channel = self.thermistor_info.pop("channel", self.valve)
+            self.thermistor_info["close_valve"] = close_on_active
             self.thermistor = ThermistorHandler(
                 self,
                 channel=thermistor_channel,
@@ -294,10 +304,11 @@ class ValveHandler:
         await asyncio.sleep(timeout)
         await self.finish(did_timeout=True)
 
-    async def finish(self, did_timeout: bool = False):
+    async def finish(self, close_valve: bool = True, did_timeout: bool = False):
         """Finishes the fill, closing the valve."""
 
-        await self._set_state(False, did_timeout=did_timeout)
+        if close_valve:
+            await self._set_state(False, did_timeout=did_timeout)
 
         self._monitor_task = await cancel_task(self._monitor_task)
         self._timeout_task = await cancel_task(self._timeout_task)
